// A. Sints
#include "stdafx.h"

#include "..\StukaDlg.h"
#include "FFMPEGCapture.h"


// input
#include "SInput.h"


// needed for ffmpeg
#ifndef INT64_C
	#define INT64_C(c) (c ## LL)
	#define UINT64_C(c) (c ## ULL)
#endif


// ffmpeg
extern "C" {
	#include "libavcodec\avcodec.h"
	#include "libavformat\avformat.h"
	#include "libswscale\swscale.h"
	#include "libavutil\opt.h"
}


// ffmpeg library
#pragma comment( lib, "avcodec.lib" )
#pragma comment( lib, "avformat.lib" )
#pragma comment( lib, "avutil.lib" )
#pragma comment( lib, "swscale.lib" )


// interpolation type by default
int CFFMPEGCapture::m_iInterpolationType = SWS_FAST_BILINEAR;

// number of image to buffer in advance
const DWORD CFFMPEGCapture::m_dwImageBuffering = 8;


// dflt ctor
CFFMPEGCapture::CFFMPEGCapture( const CSInput & _rInput, const DWORD _dwFrameSeekMs, const DWORD _dwNbFrameToCapture, const DWORD _dwOutWidth, const DWORD _dwOutHeight, const DWORD _dwDisplayFrame, const DWORD _dwDNRLevel, const CSIAdjustment::eFlip _eVerticalFlip )
	: m_rInput( _rInput ), m_dwFrameSeekMs( _dwFrameSeekMs ), m_dwNbFrameToCapture( _dwNbFrameToCapture ),
	m_dwOutWidth( _dwOutWidth ), m_dwOutHeight( _dwOutHeight ), m_dwDisplayFrame( _dwDisplayFrame ),
	m_bCaptureError( false ), m_hBmpNULL( NULL ), m_dwCurrentFrame( 0 ), m_pu8DNRBuffer( NULL ),
	m_dwDNRLevel( _dwDNRLevel ), m_eVerticalFlip( _eVerticalFlip ), m_dwLastGetBitmap( 0 ) {

	// not started
	m_evtStarted.ResetEvent();
}



// dtor
CFFMPEGCapture::~CFFMPEGCapture() {

	// stop the thread first
	CThread::Stop();

	// release remaining bitmaps
	CSingleLock lock( &m_mtxBmpMap, TRUE );
	for( POSITION pos = m_mBmp.GetStartPosition(); pos; ) {
		DWORD dwPosition;
		HBITMAP hBmp;
		m_mBmp.GetNextAssoc( pos, dwPosition, hBmp );
		DeleteObject( hBmp );
	}

	// delete DNR buffer
	delete [] m_pu8DNRBuffer;
}


// start capture
const bool CFFMPEGCapture::StartCapture() {

	// just start the thread
	CString sThreadName;
	sThreadName.Format( L"FFMPEGCapture(%s)", m_rInput.GetID().GetString() );
	bool bRes = CThread::Start( sThreadName );

	// wait for thread
	WaitForSingleObject( m_evtStarted, INFINITE );
	Sleep( 1 );
	return bRes;
}


// capture frames
const bool CFFMPEGCapture::_Capture() {

//L_Capture:

	// lock manager
	if( av_lockmgr_register( NULL ) != 0 ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to register lock manager.\n" );
		m_evtStarted.SetEvent();
		return false;
	}

	// register all formats
	av_register_all();

	// open video file
	AVFormatContext * pFormatCtx;
	if( av_open_input_file( &pFormatCtx, CStringA( m_rInput.GetVideoPath() ).GetString(), NULL, 0, NULL ) != 0 ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to open video file.\n" );
		m_evtStarted.SetEvent();
		return false;
	}

	// find stream
	if( av_find_stream_info( pFormatCtx ) < 0 ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Couldn't find stream information.\n" );
		m_evtStarted.SetEvent();
		return false;
	}

	// find the first video stream, width and height
	int iVideoStream = -1;
	DWORD dwSrcWidth = 0, dwSrcHeight = 0;
	for( int i = 0; i < int( pFormatCtx->nb_streams ) && iVideoStream == -1; i++ ) {
		if( pFormatCtx->streams[ i ]->codec->codec_type == CODEC_TYPE_VIDEO ) {
			iVideoStream = i;
			dwSrcWidth = pFormatCtx->streams[ i ]->codec->width;
			dwSrcHeight = pFormatCtx->streams[ i ]->codec->height;
		}
	}
	if( iVideoStream == -1 ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Didn't find a video stream.\n" );
		m_evtStarted.SetEvent();
		return false;
	}

	// get a pointer to the codec context for the video stream
	AVCodecContext * pCodecCtx = pFormatCtx->streams[ iVideoStream ]->codec;

	// find the decoder for the video stream
	AVCodec * pCodec = avcodec_find_decoder( pCodecCtx->codec_id );
	if( pCodec == NULL ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Unsupported codec.\n" );
		m_evtStarted.SetEvent();
		return false;
	}

	// open codec
	if( avcodec_open( pCodecCtx, pCodec ) < 0 ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to open codec.\n" );
		m_evtStarted.SetEvent();
		return false;
	}

	// started
	m_evtStarted.SetEvent();

	// allocate video frame
	AVFrame * pFrame = avcodec_alloc_frame();
	if( pFrame == NULL ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to allocate frame.\n" );
		return false;
	}

	// allocate an AVFrame structure
	AVFrame * pFrameRGB = avcodec_alloc_frame();
	if( pFrameRGB == NULL ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to allocate frame RGB.\n" );
		return false;
	}

	// determine required buffer size and allocate buffer
	int numBytes = avpicture_get_size( PIX_FMT_BGR24, m_dwOutWidth, m_dwOutHeight );
	uint8_t * buffer = ( uint8_t * ) av_malloc( numBytes * sizeof( uint8_t ) );

	// assign appropriate parts of buffer to image planes in pFrameRGB
	if( avpicture_fill( ( AVPicture * ) pFrameRGB, buffer, PIX_FMT_BGR24, m_dwOutWidth, m_dwOutHeight ) < 0 ) {
		CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to fill buffer.\n" );
		return false;
	}

	// time between two input/output frames (ms)
	double dTB2FIn = double( pFormatCtx->streams[ iVideoStream ]->r_frame_rate.den * 1000 ) / pFormatCtx->streams[ iVideoStream ]->r_frame_rate.num;
	double dTB2FOut = 40; // 1000 / 25

	// seek position
	DWORD dwSeekPosFrame = ( m_dwFrameSeekMs * pFormatCtx->streams[ iVideoStream ]->r_frame_rate.num ) / ( pFormatCtx->streams[ iVideoStream ]->r_frame_rate.den * 1000 );

	// try to seek the file
	DWORD dwKeyIntMin = DWORD( pCodecCtx->keyint_min * 3 );
	if( dwSeekPosFrame > dwKeyIntMin ) {
		DWORD dwMult = pFormatCtx->streams[ iVideoStream ]->time_base.den / pFormatCtx->streams[ iVideoStream ]->r_frame_rate.num;
		int64_t dwSeekReached = int64_t( dwSeekPosFrame - dwKeyIntMin ) * dwMult;
		if( avformat_seek_file( pFormatCtx, iVideoStream, 0, dwSeekReached, dwSeekReached, AVSEEK_FLAG_ANY ) >= 0 ) {
			avcodec_flush_buffers( pFormatCtx->streams[ iVideoStream ]->codec );
			dwSeekPosFrame = dwKeyIntMin;
		}
	}
	
	// compute cropping rectangle
	CRect rctCrop = _ComputeCropRect( dwSrcWidth, dwSrcHeight );
	
	// read a frame
	bool bRes = false;

	// source frame count
	int iSrcFrame = -1;

	// for each frame to capture
	// we need to skip/copy frame depending on src/dst FPS
	for( m_dwCurrentFrame = 0; m_dwCurrentFrame <= m_dwNbFrameToCapture && !CThread::Stopping(); m_dwCurrentFrame++ ) {

		// we need to buffer only a certain number of frames
		while( ( m_dwCurrentFrame > m_dwLastGetBitmap + m_dwImageBuffering ) && !CThread::Stopping() )
			Sleep( 1 );
		if( CThread::Stopping() )
			break;

		// reset return value
		bRes = false;

		// compute desired frame
		int iDesiredFrame = int( double( m_dwCurrentFrame ) * dTB2FOut / dTB2FIn );

		// reuse the last buffer if we ask for the same frame
		if( iSrcFrame != iDesiredFrame ) {

			// search for desired video frame
			AVPacket packet;
			bool bSubRes = false;
			while( iSrcFrame != iDesiredFrame && !CThread::Stopping() ) {

				bool bVideoFound = false;
				while( av_read_frame( pFormatCtx, &packet ) >= 0 && !CThread::Stopping() ) {
				
					// is this a packet from the video stream?
					if( packet.stream_index == iVideoStream ) {
						bVideoFound = true;
						break;
					}
					else			
						// release packet
						av_free_packet( &packet );
				}

				// video not found
				if( !bVideoFound ) {
					CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to decode frame.\n" );
					break;
				}

				// skip until desired frame
				if( dwSeekPosFrame > 0 )
					dwSeekPosFrame--;
				else
					// next src frame
					iSrcFrame++;

				// decode video frame, we can skip frame decoding while seeking
				int iFrameFinished = 1;
				// by security we triple the keyframe interval
				if( dwSeekPosFrame <= dwKeyIntMin ) {
					if( avcodec_decode_video( pCodecCtx, pFrame, &iFrameFinished, packet.data, packet.size ) < 0 ) {
						CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to decode video.\n" );
						break;
					}
				}

				// free packet
				av_free_packet( &packet );

				// did we get a video frame?
				if( iFrameFinished == 32 ) { // == 0 ?
					CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to decode frame.\n" );
					break;
				}

				// ok
				bSubRes = true;
			}

			// failure
			if( !bSubRes )
				break;
			
			// convert the image from its native format to RGB
			struct SwsContext * img_convert_ctx = sws_getContext( rctCrop.Width(), rctCrop.Height(), pCodecCtx->pix_fmt, m_dwOutWidth, m_dwOutHeight, PIX_FMT_BGR24, m_iInterpolationType, NULL, NULL, NULL );
			if( img_convert_ctx == NULL ) {
				CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to get sws context.\n" );
				break;
			}

			// handle invertion			
			bool bInvert;
			if( m_eVerticalFlip == CSIAdjustment::FT_AUTO )
				// auto detect depends of the type...
				bInvert = pCodecCtx->pix_fmt == PIX_FMT_YUVJ422P ? false : true;
			else
				bInvert = ( m_eVerticalFlip == CSIAdjustment::FT_TRUE );

			// handle cropping and flip vertically
			int nMaxLineSize = 0; 
			for( int i = 0; i < 4; i++ )
				if ( pFrame->linesize[ i ] > nMaxLineSize )
					nMaxLineSize = pFrame->linesize[ i ];
			for( int i = 0; i < 4; i++ ) {
				if( pFrame->linesize[ i ] != 0 ) {
					int nDivisor = nMaxLineSize == 0 ? 1 : ( nMaxLineSize / pFrame->linesize[ i ] );
					if( bInvert ) {
						pFrame->data[ i ] += ( pFrame->linesize[ i ] * ( ( pCodecCtx->height / nDivisor ) - 1 ) )					// invert image
										- ( ( pFrame->linesize[ i ] * ( ( pCodecCtx->height - rctCrop.bottom ) / nDivisor ) )		// vertical crop
										- ( rctCrop.left / nDivisor ) );															// horizontal crop
						pFrame->linesize[ i ] *= -1; // invert image
					}
					else {
						pFrame->data[ i ] += ( ( pFrame->linesize[ i ] * ( ( pCodecCtx->height - rctCrop.bottom ) / nDivisor ) )	// vertical crop
										  + ( rctCrop.left / nDivisor ) );															// horizontal crop
					}
				} 
			}

			// scale
			int iScaleRes = sws_scale( img_convert_ctx, pFrame->data, pFrame->linesize, 0, pCodecCtx->height, pFrameRGB->data, pFrameRGB->linesize );
			sws_freeContext( img_convert_ctx );
			if( iScaleRes < 0 ) {
				CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to get sws context.\n" );
				break;
			}
		}

		// create BITMAPINFO header
		BITMAPINFO bmi;
		ZeroMemory( &bmi, sizeof( BITMAPINFO ) );
		bmi.bmiHeader.biSize = sizeof( BITMAPINFOHEADER );
		bmi.bmiHeader.biBitCount = 24;
		bmi.bmiHeader.biWidth = m_dwOutWidth;
		bmi.bmiHeader.biHeight = m_dwOutHeight;
		bmi.bmiHeader.biPlanes = 1;

		// create DIB Section
		HBITMAP hBmp = CreateDIBSection( NULL, &bmi, DIB_RGB_COLORS, m_dwDNRLevel != 0 ? NULL : ( void ** )&( pFrameRGB->data ), NULL, 0 );
		if( hBmp == NULL ) {
			CStukaDlg::Output( L"Error (FFMPEGCapture): Failed to create DIB section.\n" );
			break;
		}

		// apply DNR
		if( m_dwDNRLevel != 0 )
			_DynamicNoiseReduction( ( unsigned char * ) *( ( void ** )&( pFrameRGB->data ) ), hBmp, m_dwDNRLevel );

		// add bmp in map
		CSingleLock lock( &m_mtxBmpMap, TRUE );
		m_mBmp.SetAt( m_dwCurrentFrame, hBmp );
		
		// all is ok
		bRes = true;
	}

	// release data
	av_free( buffer );
	av_free( pFrameRGB );
	av_free( pFrame );
	avcodec_close( pCodecCtx );
	av_close_input_file( pFormatCtx );

	// return result
	return bRes;
}


// callback
const bool CFFMPEGCapture::Callback() {

	// capture
	m_bCaptureError = !_Capture();

	// stop callback loop when done
	return false;
}


// get bitmap
const HBITMAP & CFFMPEGCapture::GetBitmap( DWORD _dwPosition ) {

	// sleep a little, let FFMPEG have a breathe
	const DWORD dwRetryDelay = 100;

	// wrong position
	if( _dwPosition >= m_dwNbFrameToCapture ) {
		CStukaDlg::Output( L"Warning (FFMPEGCapture): Wrong GetBitmap position (over the end of the segment).\n" );
		_dwPosition = m_dwNbFrameToCapture - 1;
	}

	// capture error (no wait)
	if( m_bCaptureError )
		return m_hBmpNULL;

	// frame not ready
	if( _dwPosition + 1 >= m_dwCurrentFrame ) {
		Sleep( dwRetryDelay );
		return m_hBmpNULL;
	}

	// get bitmap
	const CMap< DWORD, DWORD, HBITMAP, HBITMAP >::CPair * pPair = NULL;
	{
		CSingleLock lock( &m_mtxBmpMap, TRUE );
		pPair = m_mBmp.PLookup( _dwPosition );
	}

	// not already in map we wait a little
	if( pPair == NULL ) {
		Sleep( dwRetryDelay );
		return m_hBmpNULL;
	}
	else {
		// return bitmap
		m_dwLastGetBitmap = _dwPosition;
		return pPair->value;
	}
}


// apply some dynamic noise reduction to a given bitmap
void CFFMPEGCapture::_DynamicNoiseReduction( const unsigned char * _pu8Src, HBITMAP & _hDstBmp, const DWORD _dwDNRLevel ) {

	// bitmap lenght
	DWORD dwLen = m_dwOutWidth * m_dwOutHeight * 3;

	// first time
	if( m_pu8DNRBuffer == NULL ) {
	
		// set DNR buffer
		m_pu8DNRBuffer = new unsigned char[ dwLen ];
		memset( m_pu8DNRBuffer, 0, dwLen );
	
		// build divTable
		for( int i = 0; i < 766; i ++ )
			m_divTable[ i ] = unsigned char( i / 3 );
	}

	// create DIB Section for dst
	BITMAP bmpDst;
	GetObject( _hDstBmp, sizeof( BITMAP ), &bmpDst );
	unsigned char * pu8Dst = ( unsigned char * ) bmpDst.bmBits;

	// Dynamic Noise Reduction
	unsigned char ap, bp, cp;
	long lDiff;
	long lLevel = _dwDNRLevel * 255 / 100;
	long lHalfLevel = lLevel >> 1;

	// for each point
	for( DWORD dwOffset = 0; dwOffset < dwLen; ) {
	
		// for R, G, B components
		for( int i = 0; i < 3; i ++ ) {
			ap = _pu8Src[ dwOffset ];
			bp = m_pu8DNRBuffer[ dwOffset ];
			cp = ap;
			lDiff = abs( ap - bp );
			if( lDiff < lLevel )
				cp = ( lDiff > lHalfLevel ) ? m_divTable[ ap + ap + bp ] : bp;
			m_pu8DNRBuffer[ dwOffset ] = cp;
			dwOffset++;
		}
	}

	// copy current buffer in dest bitmap
	memcpy( pu8Dst, m_pu8DNRBuffer, m_dwOutWidth * m_dwOutHeight * 3 );
}


// release a given bitmap
void CFFMPEGCapture::ReleaseBitmap( const DWORD _dwPosition ) {

	// get bitmap
	CSingleLock lock( &m_mtxBmpMap, TRUE );
	const CMap< DWORD, DWORD, HBITMAP, HBITMAP >::CPair * pPair = m_mBmp.PLookup( _dwPosition );
	if( pPair != NULL ) {
		DeleteObject( pPair->value );
		m_mBmp.RemoveKey( _dwPosition );
	}
}


// compute cropping rectangle
CRect CFFMPEGCapture::_ComputeCropRect( const DWORD _dwSrcWidth, const DWORD _dwSrcHeight ) {

	// max width/height
	int iMaxWidth = int( _dwSrcWidth - 1 );
	int iMaxHeight = int( _dwSrcHeight - 1 );

	// 0% == maximize and deform
	if( m_dwDisplayFrame == 0 )
		return ( m_rctCapture.Width() == 0 ) ? CRect( CPoint( 0, 0 ), CPoint( iMaxWidth, iMaxHeight ) ) : m_rctCapture;

	// get focus point
	CPoint ptFocus = const_cast< CSInput & >( m_rInput ).BuildFocusPoint( _dwSrcWidth, _dwSrcHeight );

	// compute best out width/height
	DWORD dwFOutWidth, dwFOutHeight;
	DWORD dwOutWidthInSrcRef = m_dwOutWidth * _dwSrcHeight / m_dwOutHeight;
	if( dwOutWidthInSrcRef <= _dwSrcWidth ) {
		dwFOutHeight = ( _dwSrcHeight * m_dwDisplayFrame / 100 ) >> 1;
		dwFOutWidth = ( ( m_dwOutWidth * _dwSrcHeight * m_dwDisplayFrame ) / ( m_dwOutHeight * 100 ) ) >> 1;
	}
	else {
		dwFOutWidth = ( _dwSrcWidth * m_dwDisplayFrame / 100 ) >> 1;
		dwFOutHeight = ( ( m_dwOutHeight * _dwSrcWidth * m_dwDisplayFrame ) / ( m_dwOutWidth * 100 ) ) >> 1;
	}

	// focus may be un-centered
	CRect rctCrop( CPoint( ptFocus.x - dwFOutWidth, ptFocus.y - dwFOutHeight ), CPoint( ptFocus.x + dwFOutWidth, ptFocus.y + dwFOutHeight ) );
	if( rctCrop.left < 0 ) rctCrop.OffsetRect( CSize( -rctCrop.left, 0 ) );
	if( rctCrop.top < 0 ) rctCrop.OffsetRect( CSize( 0, -rctCrop.top ) );
	if( rctCrop.right > iMaxWidth ) rctCrop.OffsetRect( iMaxWidth - rctCrop.right, 0 );
	if( rctCrop.bottom > iMaxHeight ) rctCrop.OffsetRect( 0, iMaxHeight - rctCrop.bottom );

	// hard bounding of cropping rectangle
	if( rctCrop.left < 0 ) rctCrop.left = 0;
	if( rctCrop.top < 0 ) rctCrop.top = 0;
	if( rctCrop.right > iMaxWidth ) rctCrop.right = _dwSrcWidth - 1;
	if( rctCrop.bottom > iMaxHeight ) rctCrop.bottom = _dwSrcHeight - 1;

	// return cropping rectangle
	return rctCrop;
}