// A. Sints
#include "stdafx.h"

#include "..\StukaDlg.h"
#include "Project.h"

// math
#include <math.h>

// AVI Writer
#include "AviWriter.h"

// marker
#include "SMarker.h"

// FFMPEG Capture
#include "FFMPEGCapture.h"

// RGB Adjustment
#include "RGBAdjustment.h"

// Layout
#include "Layout.h"

// input
#include "SInput.h"

// output dlg
#include "../OutputDlg.h"


// dflt ctor
CProject::CProject( LPCTSTR _sScriptPath, const COutputDlg & _rOutputDlg )
	: m_script( _sScriptPath ), m_rOutputDlg( _rOutputDlg ) {

	// just start the thread
	CThread::Start( L"StukaProject" );

	// create text DC
	m_textDC.CreateCompatibleDC( NULL );

	// set text font
	LOGFONT logFont;
	m_textDC.GetCurrentFont()->GetLogFont( &logFont );
	lstrcpy( logFont.lfFaceName, L"Calibri" );
	logFont.lfQuality			= CLEARTYPE_QUALITY;
	logFont.lfPitchAndFamily	= DEFAULT_PITCH | FF_SWISS;
	logFont.lfClipPrecision		= CLIP_DEFAULT_PRECIS;
	logFont.lfWidth				= 0;
	logFont.lfHeight			= 20;
	logFont.lfWeight			= FW_MEDIUM;
	logFont.lfItalic			= FALSE;
	m_font.CreateFontIndirect( &logFont );
}


// dtor
CProject::~CProject() {

	// stop
	CThread::Stop();
}


// callback
const bool CProject::Callback() {

	// build
	if( !Build() )
		CStukaDlg::Output( L"Error (Project): Failed to build project.\n" );
	else
		CStukaDlg::Output( L"Done without error.\n" );
	CStukaDlg::EndBuild();

	// stop callback loop when done
	return false;
}


// build project
const bool CProject::Build() {

	// timer
	double t0 = _GetTime();
	CStukaDlg::Progress( 0 );
	CStukaDlg::Progress( 0, CStukaDlg::PT_TOTAL );
	CStukaDlg::Remaining( L"" );

	CStukaDlg::Output( L"* Parsing script...\n" );

	// parse script
	if( !m_script.Parse() ) {
		CStukaDlg::Output( L"Error (Project): Failed to parse script.\n" );
		return false;
	}	

	// validate script
	if( !m_script.Validate() ) {
		CStukaDlg::Output( L"Error (Project): Failed to validate script.\n" );
		return false;
	}

	// create AVI Writer
	CAVIWriter AVIWriter( m_script.GetOutputPath(), m_script.GetOutputWidth(), m_script.GetOutputHeight(), m_script.GetAudioPath() );
	if( !AVIWriter.Create() ) {
		CStukaDlg::Output( L"Error (Project): Failed to create AVI Writer.\n" );
		return false;
	}

	// multiply overlay
	CFFMPEGCapture * pFFMPEGCaptureOverlay = NULL;
	DWORD dwOverlayCount = 0;
	DWORD dwOverlayFrame = 0;
	CSInput * pOverlayInput = NULL;
	if( !m_script.GetOverlay().GetRefID().IsEmpty() ) {
	
		CStukaDlg::Output( L"* Building overlay frames...\n" );

		// get overlay input
		CList< CSInput * > lInputs;
		if( !m_script.BuildInputList( m_script.GetOverlay().GetRefID(), lInputs ) ) {
			CStukaDlg::Output( L"Error (Project): Failed to build input list (overlay).\n" );
			return false;
		}
		pOverlayInput = lInputs.GetHead();

		// create FFMPEGCapture
		dwOverlayCount = m_script.GetOverlay().GetLoopCount();
		pFFMPEGCaptureOverlay = new CFFMPEGCapture( *pOverlayInput, 0, dwOverlayCount, m_script.GetOutputWidth(), m_script.GetOutputHeight(), 0, pOverlayInput->GetRGBAdjustment().GetDNRLevel(), pOverlayInput->GetRGBAdjustment().GetVerticalFlip() );

		// set capture rect
		pFFMPEGCaptureOverlay->SetCaptureRect( m_script.GetOverlay().GetCaptureRect() );

		// start overlay capture
		if( !pFFMPEGCaptureOverlay->StartCapture() ) {
			CStukaDlg::Output( L"Error (Project): Failed to start overlay capture.\n" );
			return false;
		}

		// loop until ok
		CString sBlock;
		sBlock.Format( L"[Computing %i frames]\n", dwOverlayCount );
		CStukaDlg::Output( sBlock.GetString() );
		for( DWORD dwFrame = 0; dwFrame < dwOverlayCount; ) {
			if( pFFMPEGCaptureOverlay->GetBitmap( dwFrame ) != NULL ) {
				CStukaDlg::Progress( unsigned char( ( dwFrame + 1 ) * 100 / DWORD( dwOverlayCount ) ) );
				dwFrame ++;
			}
			if( pFFMPEGCaptureOverlay->CaptureError() ) {
				CStukaDlg::Output( L"Error (Project): Failed to capture overlay source video frames.\n" );
				return false;
			}
		}
	}

	CStukaDlg::Progress( 0 );

	// we need to estimate total frame count
	DWORD dwTempoEst = 0;
	CSMSignature signatureEst;
	double dTotalMs = 0;
	for( POSITION pos = m_script.GetMarkerList().GetHeadPosition(); pos; ) {
		const CSMarker & rMarkerA = *( m_script.GetMarkerList().GetNext( pos ) );
		const CSMarker & rMarkerB = *( m_script.GetMarkerList().GetAt( pos ) );
		const DWORD & dwPosA = rMarkerA.GetPosition();
		const DWORD & dwPosB = rMarkerB.GetPosition();

		// do we need to change tempo or signature ?
		if( rMarkerA.GetTempo() != 0 )
			dwTempoEst = rMarkerA.GetTempo();
		if( rMarkerA.GetSignature().GetBeatCount() != 0 )
			signatureEst = rMarkerA.GetSignature();

		// determine total ms
		dTotalMs += double( ( ( dwPosB - dwPosA ) * signatureEst.GetBeatCount() * 60000 ) / ( double( signatureEst.GetBeatUnit() * dwTempoEst ) / 4 ) );
		
		// we've done the job
		if( rMarkerB.IsTimelineEnd() )
			break;
	}

	double t1 = _GetTime();
	double tEst = 0;

	// total frame (25fps output)
	DWORD dwTotalFrameCount = DWORD( dTotalMs * 25 / 1000 );

	// get writer data
	BITMAP bmpWriter;
	GetObject( AVIWriter.GetBitmap(), sizeof( BITMAP ), &bmpWriter );
	unsigned char * pu8Dst = ( unsigned char * ) bmpWriter.bmBits;

	// current tempo and signature
	DWORD dwTempo = 0;
	DWORD dwSeparatorSize = 0;
	CSMSignature signature;
	double dCurrPositionMs = 0;
	DWORD dwCurrentFrame = 0;

	// for each markers
	for( POSITION pos = m_script.GetMarkerList().GetHeadPosition(); pos && !CThread::Stopping(); ) {
		const CSMarker & rMarkerA = *( m_script.GetMarkerList().GetNext( pos ) );
		const CSMarker & rMarkerB = *( m_script.GetMarkerList().GetAt( pos ) );
		
		// do we need to change tempo ?
		if( rMarkerA.GetTempo() != 0 )
			dwTempo = rMarkerA.GetTempo();

		// do we need to change signature ?
		if( rMarkerA.GetSignature().GetBeatCount() != 0 )
			signature = rMarkerA.GetSignature();

		// do we need to change separator size
		if( rMarkerA.GetSeparatorSize() != -1 )
			dwSeparatorSize = DWORD( rMarkerA.GetSeparatorSize() );

		// get marker positions
		const DWORD & dwPosA = rMarkerA.GetPosition();
		const DWORD & dwPosB = rMarkerB.GetPosition();

		// stop ?
		const DWORD & dwRangeEnd = m_script.GetRangeEnd();
		if( dwRangeEnd != 0 && dwPosA == dwRangeEnd )
			break;

		// building timeline A to B
		CString sOut;
		DWORD dwEndTag = dwRangeEnd != 0 ? dwRangeEnd : m_script.GetMarkerList().GetTail()->GetPosition();
		sOut.Format( L"* Building timeline markers %i to %i (end@%i) - %i/%i @ %ibpm\n", dwPosA, dwPosB, dwEndTag, signature.GetBeatCount(), signature.GetBeatUnit(), dwTempo );
		CStukaDlg::Output( sOut.GetString() );

		// get layout
		CLayout layout( m_script.GetOutputWidth(), m_script.GetOutputHeight() );
		CList< CRect > lLayoutRects;
		CList< CRect > lLayoutLines;
		if( !layout.GetLayout( rMarkerA.GetLayoutType(), lLayoutRects, lLayoutLines ) ) {
			CStukaDlg::Output( L"Error (Project): Failed to get layout.\n" );
			return false;
		}

		// viewport list
		CList< CViewport * > lViewports;

		// we need to check that the layout count is equal to the display count
		if( lLayoutRects.GetCount() != rMarkerA.GetDisplayList().GetCount() ) {
			CStukaDlg::Output( L"Error (Project): Display list count does not correspound to the layout rectangles count.\n" );
			return false;
		}

		// determine min cycle time
		double dBUTem			= double( signature.GetBeatUnit() * dwTempo ) / 4;
		double dEveryEighthBeatMs	= double( 60000 / ( dBUTem * 8 ) );
		double dEveryQuarterBeatMs	= double( 60000 / ( dBUTem * 4 ) );
		double dEveryHalfBeatMs		= double( 60000 / ( dBUTem * 2 ) );
		double dEachBeatMs			= double( 60000 / dBUTem );
		double dEvery2BeatMs		= double( 2 * 60000 / dBUTem );
		double dEvery3BeatMs		= double( 3 * 60000 / dBUTem );
		double dEachMeasureMs		= double( signature.GetBeatCount() * 60000 / dBUTem );
		double dMarkerMs			= double( ( ( dwPosB - dwPosA ) * signature.GetBeatCount() * 60000 ) / dBUTem );

		// position to layout
		POSITION posLayout = lLayoutRects.GetHeadPosition();

		//  get min cycle time
		double dMinCycleMs = dMarkerMs;

		// feed layout with display in order
		for( POSITION posD = rMarkerA.GetDisplayList().GetHeadPosition(); posD; ) {
			const CSMDisplay & rDisplay = *( rMarkerA.GetDisplayList().GetNext( posD ) );

			// add a viewport
			CViewport & rViewport = *( new CViewport( rDisplay.GetRefID(), lLayoutRects.GetNext( posLayout ) ) );
			lViewports.AddTail( &rViewport );
			
			// use script to get the full list of inputs
			if( !m_script.BuildInputList( rDisplay.GetRefID(), rViewport.GetInputList() ) ) {
				CStukaDlg::Output( L"Error (Project): Failed to build input list.\n" );
				return false;
			}

			// determine current cycle
			double dCurrentCycleMs = _GetCycleTime( rDisplay.GetCycle(), dMinCycleMs, dEachMeasureMs, dEvery3BeatMs, dEvery2BeatMs, dEachBeatMs, dEveryHalfBeatMs, dEveryQuarterBeatMs, dEveryEighthBeatMs );
			
			// get min cycle time
			if( dCurrentCycleMs < dMinCycleMs ) dMinCycleMs = dCurrentCycleMs;

			// odd measures...
			while( 1 ) {
				double dOddCycle = dMarkerMs - double( DWORD( dMarkerMs / dMinCycleMs ) ) * dMinCycleMs;
				if( dOddCycle < 0.0000001 )
					break;
				if( dOddCycle < dMinCycleMs )
					dMinCycleMs = dOddCycle;
			}

			// set viewport cycle
			rViewport.SetCycle( dCurrentCycleMs );
		}

		// 25 fps
		double dBlockCount = dMarkerMs / dMinCycleMs;
		double dFramesPerBlock = dMinCycleMs * 25 / 1000;
		sOut.Format( L"%i block(s) of %ims (%i frames) to compute\n", DWORD( dBlockCount ), DWORD( dMinCycleMs ), DWORD( dFramesPerBlock ) );
		CStukaDlg::Output( sOut.GetString() );

		// skip
		const DWORD & dwRangeBegin = m_script.GetRangeBegin();
		if( dwRangeBegin != 0 && dwPosA < dwRangeBegin ) {
			
			// next cycle
			dwCurrentFrame += ( DWORD( dFramesPerBlock ) * DWORD( dBlockCount ) );

			// offset writer
			AVIWriter.Offset( dwCurrentFrame );

			// release viewports
			for( POSITION posV = lViewports.GetHeadPosition(); posV; )
				delete lViewports.GetNext( posV );
			lViewports.RemoveAll();

			// increase current position
			dCurrPositionMs += dMarkerMs;

			// we've done the job (just in case...)
			if( rMarkerB.IsTimelineEnd() )
				break;

			// skip
			continue;
		}
		
		// for each block
		double dCurrentCycleMs = 0;
		for( DWORD i = 0; i < DWORD( dBlockCount ) && !CThread::Stopping(); i++ ) {
			
			// prepare viewport viewports
			for( POSITION posV = lViewports.GetHeadPosition(); posV; ) {
				CViewport & rViewport = *( lViewports.GetNext( posV ) );
				rViewport.Cycle( dCurrentCycleMs );
				CSInput & rInput = rViewport.GetCurrentInput();
				// seek position
				DWORD dwFrameSeekMs = DWORD( dCurrPositionMs + dCurrentCycleMs + double( rInput.GetAnchorDifference() ) );
				// frames to capture depends of the speed value
				DWORD dwNbFrameToCapture = DWORD( dFramesPerBlock * rMarkerA.GetSpeedValue( rInput.GetID() ) ) + 1;
				CFFMPEGCapture * pFFMPEGCapture = new CFFMPEGCapture( rInput, dwFrameSeekMs, dwNbFrameToCapture, rViewport.GetLayout().Width(), rViewport.GetLayout().Height(), rMarkerA.GetResizeValue( rInput.GetID() ), rInput.GetRGBAdjustment().GetDNRLevel(), rInput.GetRGBAdjustment().GetVerticalFlip() );
				rViewport.SetFFMPEGCapture( *pFFMPEGCapture );
			}

			// start capture
			for( POSITION posV = lViewports.GetHeadPosition(); posV; ) {
				if( !lViewports.GetNext( posV )->GetFFMPEGCapture().StartCapture() ) {
					CStukaDlg::Output( L"Error (Project): Failed to start capture.\n" );
					return false;
				}
			}

			// reset progress for current block
			CString sBlock;
			sBlock.Format( L"[Block %i/%i]\n", i + 1, DWORD( dBlockCount ) );
			CStukaDlg::Output( sBlock.GetString() );
			CStukaDlg::Progress( 0 );
			
			// total block frame count
			DWORD dwTotalBlockFrame = DWORD( dBlockCount ) * DWORD( dFramesPerBlock );

			// treat all bitmaps
			for( DWORD dwFrame = 0; dwFrame < DWORD( dFramesPerBlock ) && !CThread::Stopping(); ) {

				// does all the viewports frame captured ?
				bool bViewportsFrameCaptured = true;
				for( POSITION posV = lViewports.GetHeadPosition(); posV && bViewportsFrameCaptured; ) {
					const CViewport & rViewport = *( lViewports.GetNext( posV ) );
					bViewportsFrameCaptured = ( rViewport.GetFFMPEGCapture().GetBitmap( DWORD( double( dwFrame ) / double( 1 / rMarkerA.GetSpeedValue( rViewport.GetCurrentInput().GetID() ) ) ) ) != NULL );
				}

				// we can now insert the frame in output
				if( bViewportsFrameCaptured ) {

					// get current effect
					unsigned char u8MulFactor = 255;
					const CSMEffect::eEffect & eEffect = rMarkerA.GetEffect().GetEffect();

					// fade in / fade out
					DWORD dwCurrentFrameInBlock = i * DWORD( dFramesPerBlock ) + dwFrame;
					DWORD dwTotalBlockFrameM1 = dwTotalBlockFrame - 1;
					if( eEffect == CSMEffect::ME_FADE_IN )
						u8MulFactor = unsigned char( dwCurrentFrameInBlock * 255 / dwTotalBlockFrameM1 );
					else
					if( eEffect == CSMEffect::ME_FADE_OUT )
						u8MulFactor = unsigned char( ( dwTotalBlockFrameM1 - dwCurrentFrameInBlock ) * 255 / dwTotalBlockFrameM1 );

					// get overlay frame if exist
					HBITMAP hBmpOverlay = NULL;
					if( pFFMPEGCaptureOverlay != NULL ) {
						hBmpOverlay = pFFMPEGCaptureOverlay->GetBitmap( dwOverlayFrame );
						dwOverlayFrame ++;
						if( dwOverlayFrame == dwOverlayCount )
							dwOverlayFrame = 0;
					}

					// copy bits
					for( POSITION posV = lViewports.GetHeadPosition(); posV && !CThread::Stopping(); ) {
						const CViewport & rViewport = *( lViewports.GetNext( posV ) );
						const CRect & rctLayout = rViewport.GetLayout();
						
						// get speed rate
						double dPlayRate = rMarkerA.GetSpeedValue( rViewport.GetCurrentInput().GetID() );
						double dDiv = double( 1 / dPlayRate );
						DWORD dwRFrame = DWORD( double( dwFrame ) / dDiv );

						// get current source bitmap
						HBITMAP hSrcBmp = rViewport.GetFFMPEGCapture().GetBitmap( dwRFrame );

						// interpolate bitmap if necessary
						if( DWORD( dDiv ) != 0 && ( dwRFrame * DWORD( dDiv ) ) != dwFrame )
							_BmpInterpolate( hSrcBmp, rViewport, rctLayout, dwFrame, dwRFrame, DWORD( dDiv ) );

						// strobe effect
						if( eEffect == CSMEffect::ME_STROBE &&
							( dwRFrame % DWORD(
							_GetCycleTime( rMarkerA.GetEffect().GetStrobeCycle(), dMinCycleMs, dEachMeasureMs, dEvery3BeatMs, dEvery2BeatMs, dEachBeatMs, dEveryHalfBeatMs, dEveryQuarterBeatMs, dEveryEighthBeatMs ) * 25 / 1000 
							) ) == 0 )
							_BmpHighlight( hSrcBmp, rctLayout, 64 );

						// copy source frame (or intermediate) to output
						_BmpCopy( hSrcBmp, rctLayout.Width(), rctLayout.Height(), pu8Dst, m_script.GetOutputWidth(), rctLayout.left, rctLayout.top, rViewport.GetCurrentInput().GetRGBAdjustment().GetTable(), u8MulFactor, hBmpOverlay, hBmpOverlay == NULL ? NULL : pOverlayInput->GetRGBAdjustment().GetTable() );
						
						// release bitmap(s)
						if( int( dwRFrame ) - 1 >= 0 ) {
							const DWORD & dwLastReleasedFrameIndex = rViewport.GetLastReleasedFrameIndex();
							for( DWORD dwReleaseFrame = dwLastReleasedFrameIndex; dwReleaseFrame <= dwRFrame - 1; dwReleaseFrame ++ )
								rViewport.GetFFMPEGCapture().ReleaseBitmap( dwReleaseFrame );
							rViewport.SetLastReleasedFrameIndex( dwRFrame );
						}

						// debug mode
						if( m_script.IsDebug() ) {
							// draw text on each frame
							const CString & sRefID = rViewport.GetRefID();
							const CString & sCurID = rViewport.GetCurrentInput().GetID();
							CString sTextID;
							sTextID.Format( L"%s: %s", sRefID.GetString(), sCurID.GetString() );
							_BmpText( AVIWriter.GetBitmap(), CRect( rctLayout.left, m_script.GetOutputHeight() - rctLayout.bottom, rctLayout.right, m_script.GetOutputHeight() - rctLayout.top ), sRefID == sCurID ? sRefID.GetString() : sTextID.GetString() );
						}
					}

					// draw layout separation (black lines)
					if( dwSeparatorSize != 0 ) {
						for( POSITION posL = lLayoutLines.GetHeadPosition(); posL; ) {
							const CRect & rctLine = lLayoutLines.GetNext( posL );
							int iHeight = rctLine.Height();
							_DrawLine( rctLine.TopLeft(), iHeight == 0 ? LT_HORIZONTAL : LT_VERTICAL, iHeight == 0 ? rctLine.Width() : rctLine.Height(), pu8Dst, m_script.GetOutputWidth(), m_script.GetOutputHeight(), dwSeparatorSize );
						}
					}

					// debug mode
					if( m_script.IsDebug() ) {
						// draw marker info
						CString sMarkerID;
						// #12457 - marker( 3 ) { 7/8 @ 75 } - [marker 4 - beat 1]
						DWORD dwCurrFrame = dwCurrentFrame + dwFrame;
						double dCurrFrameMs = ( ( double( dwFrame ) + ( dFramesPerBlock * i ) ) * 1000 ) / 25;
						sMarkerID.Format( L"#%i - marker( %i ) { %i/%i @ %i } - [marker %i - beat %i]",
							dwCurrFrame + 1, rMarkerA.GetPosition(), signature.GetBeatCount(), signature.GetBeatUnit(), dwTempo,
							DWORD( dCurrFrameMs / dEachMeasureMs ) + rMarkerA.GetPosition(), 
							( DWORD( dCurrFrameMs / dEachBeatMs ) % signature.GetBeatCount() ) + 1 );
						_BmpText( AVIWriter.GetBitmap(), layout.GetRect(), sMarkerID, DT_CENTER | DT_BOTTOM );
					}

					// output
					const_cast< COutputDlg & >( m_rOutputDlg ).Output( pu8Dst, m_script.GetOutputWidth(), m_script.GetOutputHeight() );

					// insert frame in dst AVI
					if( !AVIWriter.InsertFrame( dwCurrentFrame + dwFrame ) ) {
						CStukaDlg::Output( L"Error (Project): Failed to insert frame.\n" );
						return false;
					}
					else {
						CStukaDlg::Progress( unsigned char( ( dwFrame + 1 ) * 100 / DWORD( dFramesPerBlock ) ) );
						CStukaDlg::Progress( unsigned char( ( dwCurrentFrame + dwFrame + 1 ) * 100 / dwTotalFrameCount ), CStukaDlg::PT_TOTAL );

						// time passed
						double tCurr = _GetTime();
						if( tCurr - tEst > 1.0 ) {
							tEst = tCurr;
							double dPassed = tEst - t1;
							double dRemainingEst = ( ( double( dwTotalFrameCount ) * dPassed ) / double ( dwCurrentFrame + dwFrame + 1 ) ) - dPassed;
							int iMin = int( dRemainingEst ) / 60;
							int iSec = int( dRemainingEst ) - ( iMin * 60 );
							CString sEst;
							sEst.Format( L"%i min(s) %.2i", iMin, iSec );
							CStukaDlg::Remaining( sEst );
						}
					}

					// next frame
					dwFrame ++;
				}

				// capture ended with errors
				bool bViewportsCapturedOk = true;
				for( POSITION posV = lViewports.GetHeadPosition(); posV && bViewportsCapturedOk; )
					bViewportsCapturedOk = !lViewports.GetNext( posV )->GetFFMPEGCapture().CaptureError();
				if( !bViewportsCapturedOk ) {
					CStukaDlg::Output( L"Error (Project): Failed to capture source video frames.\n" );
					return false;
				}
			}

			CStukaDlg::Progress( 0 );

			// release the FFMPEGCaptures
			for( POSITION posV = lViewports.GetHeadPosition(); posV; )
				lViewports.GetNext( posV )->ReleaseFFMPEGCapture();

			// next cycle
			dCurrentCycleMs += dMinCycleMs;
			dwCurrentFrame += DWORD( dFramesPerBlock );
		}

		// release viewports
		for( POSITION posV = lViewports.GetHeadPosition(); posV; )
			delete lViewports.GetNext( posV );
		lViewports.RemoveAll();

		// increase current position
		dCurrPositionMs += dMarkerMs;
		
		// we've done the job
		if( rMarkerB.IsTimelineEnd() )
			break;
	}

	// release Overlay FFMPEG capture
	if( pFFMPEGCaptureOverlay != NULL )
		delete pFFMPEGCaptureOverlay;
	
	// close AVI Writer
	AVIWriter.Close();

	// timer
	CString sOk;
	double dTime( _GetTime() - t0 );
	int iMin = int( dTime ) / 60;
	int iSec = int( dTime ) - ( iMin * 60 );
	sOk.Format( L"Done in %f secs [%i min(s) %.2i] for %i frames.\n", dTime, iMin, iSec, dwCurrentFrame );
	CStukaDlg::Output( sOk );
	CStukaDlg::Progress( 100, CStukaDlg::PT_TOTAL );
	CStukaDlg::Remaining( L"" );

	// ok
	return true;
}



// get current time with precision as 1. sec
const double CProject::_GetTime() {

	LARGE_INTEGER liTime, liFrequency;
	QueryPerformanceCounter( &liTime );
	QueryPerformanceFrequency( &liFrequency );
	return double( liTime.QuadPart ) / double( liFrequency.QuadPart );
}


// copy a bitmap
void CProject::_BmpCopy( const HBITMAP & _hSrcBmp, const DWORD _dwSrcWidth, const DWORD _dwSrcHeight, unsigned char * _pu8Dst, const DWORD _dwDstWidth, const DWORD _dwOffsetX, const DWORD _dwOffsetY, const unsigned char * const _pRGBAdjust, const unsigned char & _u8Factor, const HBITMAP _hBmpOverlay, const unsigned char * const _pRGBAdjustOverlay ) {

	// get source data
	BITMAP bmpSrc;
	GetObject( _hSrcBmp, sizeof( BITMAP ), &bmpSrc );
	unsigned char * pu8Src = ( unsigned char * ) bmpSrc.bmBits;

	// get overlay data
	BITMAP bmpOverlay;
	unsigned char * pu8Overlay = NULL;
	if( _hBmpOverlay != NULL ) {		
		GetObject( _hBmpOverlay, sizeof( BITMAP ), &bmpOverlay );
		pu8Overlay = ( unsigned char * ) bmpOverlay.bmBits;
	}

	// copy bits
	DWORD dwOffRGB3, dwOffRGB3Off;
	DWORD dwDstOffset = ( _dwOffsetY * _dwDstWidth + _dwOffsetX ) * 3;
	DWORD dwSrcRGBWidth = _dwSrcWidth * 3;
	DWORD dwDstRGBWidth = _dwDstWidth * 3;
	DWORD dwSrcLine, dwDstLine, dwI, dwJ;

	// for each column
	for( DWORD y = 0; y < _dwSrcHeight; y ++ ) {
		dwSrcLine = dwSrcRGBWidth * y;
		dwDstLine = dwDstOffset + dwDstRGBWidth * y;

		// for each row
		for( DWORD x = 0; x < dwSrcRGBWidth; x += 3 ) {
			dwI = dwSrcLine + x;			
			dwOffRGB3 = ( ( DWORD( pu8Src[ dwI + 2 ] ) << 16 ) | ( DWORD( pu8Src[ dwI + 1 ] ) << 8 ) | pu8Src[ dwI ] ) * 3;
			dwJ = dwDstLine + x;

			// we have an oervlay
			if( _hBmpOverlay != NULL ) {
				dwOffRGB3Off = ( ( DWORD( pu8Overlay[ dwJ + 2 ] ) << 16 ) | ( DWORD( pu8Overlay[ dwJ + 1 ] ) << 8 ) | pu8Overlay[ dwJ ] ) * 3; 

				// fade in / fade out
				if( _u8Factor != 255 )
					for( DWORD i = 0; i < 3; i ++ )
						_pu8Dst[ dwJ + i ]	= unsigned char( DWORD( _pRGBAdjust[ dwOffRGB3 + i ] ) * DWORD( _pRGBAdjustOverlay[ dwOffRGB3Off + i ] ) * DWORD( _u8Factor ) / 65025 );
				else
					for( DWORD i = 0; i < 3; i ++ )
						_pu8Dst[ dwJ + i ]	= unsigned char( DWORD( _pRGBAdjust[ dwOffRGB3 + i ] ) * DWORD( _pRGBAdjustOverlay[ dwOffRGB3Off + i ] ) / 255 );
			}
			else {
				// fade in / fade out
				if( _u8Factor != 255 )
					for( DWORD i = 0; i < 3; i ++ )
						_pu8Dst[ dwJ + i ]	= unsigned char( DWORD( _pRGBAdjust[ dwOffRGB3 + i ] ) * DWORD( _u8Factor ) / 255 );
				else
					for( DWORD i = 0; i < 3; i ++ )
						_pu8Dst[ dwJ + i ]	= _pRGBAdjust[ dwOffRGB3 + i ];
			}
		}
	}
}


// draw line
void CProject::_DrawLine( const CPoint & _ptOrigin, const eLineType _eLineType, const DWORD _dwLenght, unsigned char * _pu8Dst, const DWORD _dwDstWidth, const DWORD _dwDstHeight, const DWORD _dwLineSize ) {

	// increment / ending
	DWORD dwInc = _eLineType == LT_HORIZONTAL ? 3 : _dwDstWidth * 3;
	DWORD dwEnd = _eLineType == LT_HORIZONTAL ? _dwLenght * 3 : _dwLenght * dwInc;

	// draw
	DWORD dwHalfLineSize = _dwLineSize >> 1;
	CPoint ptOrigin;
	for( int k = -int( dwHalfLineSize ); k < int( _dwLineSize - dwHalfLineSize ); k++ ) {

		// build origin
		ptOrigin = CPoint( _ptOrigin.x + ( _eLineType == LT_VERTICAL ? k : 0 ), _ptOrigin.y + ( _eLineType == LT_HORIZONTAL ? k : 0 ) );
		if( ptOrigin.x < 0 || ptOrigin.x >= int( _dwDstWidth ) ||
			ptOrigin.y < 0 || ptOrigin.y >= int( _dwDstHeight ) )
			continue;

		// offset to origin
		DWORD dwOffset = ( ptOrigin.y * _dwDstWidth + ptOrigin.x ) * 3;

		for( DWORD i = 0; i < dwEnd; i += dwInc ) {
			for( DWORD j = 0; j < 3; j ++ )
				_pu8Dst[ dwOffset + j ] = 0;
			dwOffset += dwInc;
		}
	}
};


// interpolate bitmap
void CProject::_BmpInterpolate( HBITMAP & _rhSrcBmp, const CViewport & _rViewport, const CRect & _rctLayout, const DWORD & _dwFrame, const DWORD & _dwRFrame, DWORD _dwDiv ) {

	// get source bytes
	BITMAP bmpSrc;
	GetObject( _rhSrcBmp, sizeof( BITMAP ), &bmpSrc );
	unsigned char * pu8Src = ( unsigned char * ) bmpSrc.bmBits;

	// get intermediate bytes (and redirect to src HBITMAP)
	unsigned char * bmpSrcInt = _rViewport.GetIntermediateBmpBits( _rctLayout.Width(), _rctLayout.Height(), _rhSrcBmp );

	// get next frame
	HBITMAP hSrcBmp2 = _rViewport.GetFFMPEGCapture().GetBitmap( _dwRFrame + 1 );
	while( hSrcBmp2 == NULL )
		hSrcBmp2 = _rViewport.GetFFMPEGCapture().GetBitmap( _dwRFrame + 1 );

	// next frame bytes
	BITMAP bmpSrc2;
	GetObject( hSrcBmp2, sizeof( BITMAP ), &bmpSrc2 );
	unsigned char * pu8Src2 = ( unsigned char * ) bmpSrc2.bmBits;

	// linear interpolation
	DWORD dwSrc2P = ( ( _dwFrame - ( _dwRFrame * _dwDiv ) ) * 255 ) / _dwDiv;
	DWORD dwSrc1P = 255 - dwSrc2P;							

	// copy bits
	DWORD dwSrcRGBWidth = _rctLayout.Width() * 3;
	DWORD dwSrcLine, dwI;
	for( DWORD y = 0; y < DWORD( _rctLayout.Height() ); y ++ ) {
		dwSrcLine = dwSrcRGBWidth * y;
		for( DWORD x = 0; x < dwSrcRGBWidth; x += 3 ) {
			dwI = dwSrcLine + x;
			bmpSrcInt[ dwI ]		= unsigned char( ( DWORD( pu8Src[ dwI ] ) * dwSrc1P + DWORD( pu8Src2[ dwI ] ) * dwSrc2P ) >> 8 );
			bmpSrcInt[ dwI + 1 ]	= unsigned char( ( DWORD( pu8Src[ dwI + 1 ] ) * dwSrc1P + DWORD( pu8Src2[ dwI + 1 ] ) * dwSrc2P ) >> 8 );
			bmpSrcInt[ dwI + 2 ]	= unsigned char( ( DWORD( pu8Src[ dwI + 2 ] ) * dwSrc1P + DWORD( pu8Src2[ dwI + 2 ] ) * dwSrc2P ) >> 8 );
		}
	}
}


// highlight bitmap
void CProject::_BmpHighlight( const HBITMAP & _rhSrcBmp, const CRect & _rctLayout, const DWORD _dwHighlight ) {

	// get source bytes
	BITMAP bmpSrc;
	GetObject( _rhSrcBmp, sizeof( BITMAP ), &bmpSrc );
	unsigned char * pu8Src = ( unsigned char * ) bmpSrc.bmBits;

	// copy bits
	DWORD dwSrcRGBWidth = _rctLayout.Width() * 3;
	DWORD dwSrcLine, dwI;
	DWORD dwR, dwG, dwB;
	for( DWORD y = 0; y < DWORD( _rctLayout.Height() ); y ++ ) {
		dwSrcLine = dwSrcRGBWidth * y;
		for( DWORD x = 0; x < dwSrcRGBWidth; x += 3 ) {
			dwI = dwSrcLine + x;
			dwR = DWORD( pu8Src[ dwI ] ) + _dwHighlight;
			dwG = DWORD( pu8Src[ dwI + 1 ] ) + _dwHighlight;
			dwB = DWORD( pu8Src[ dwI + 2 ] ) + _dwHighlight;
			pu8Src[ dwI ]		= dwR > 255 ? 255 : unsigned char( dwR );
			pu8Src[ dwI + 1 ]	= dwG > 255 ? 255 : unsigned char( dwG );
			pu8Src[ dwI + 2 ]	= dwB > 255 ? 255 : unsigned char( dwB );
		}
	}
}


// get cycle time
const double CProject::_GetCycleTime( const CSMDisplay::eCycle & _reCycle, const double & _dMinCycleMs, const double & _dEachMeasureMs, const double & _dEvery3BeatMs, const double & _dEvery2BeatMs, const double & _dEachBeatMs, const double & _dEveryHalfBeatMs, const double & _dEveryQuarterBeatMs, const double & _dEveryEighthBeatMs ) {

	double dCycleMs = _dMinCycleMs;
	switch( _reCycle ) {
		case CSMDisplay::DC_NONE: break;
		case CSMDisplay::DC_EACHMEASURE:		dCycleMs = _dEachMeasureMs; break;
		case CSMDisplay::DC_EACHBEAT:			dCycleMs = _dEachBeatMs; break;
		case CSMDisplay::DC_EVERY2BEATS:		dCycleMs = _dEvery2BeatMs; break;
		case CSMDisplay::DC_EVERY3BEATS:		dCycleMs = _dEvery3BeatMs; break;
		case CSMDisplay::DC_EVERYHALFBEAT:		dCycleMs = _dEveryHalfBeatMs; break;
		case CSMDisplay::DC_EVERYQUARTERBEAT:	dCycleMs = _dEveryQuarterBeatMs; break;
		case CSMDisplay::DC_EVERYEIGHTHBEAT:	dCycleMs = _dEveryEighthBeatMs; break;
	}
	return dCycleMs;
}


// draw text on bitmap
void CProject::_BmpText( const HBITMAP & _rhSrcBmp, const CRect & _rctLayout, LPCTSTR _sText, UINT _uiFormat ) {

	// select bitmap on text DC
	m_textDC.SelectObject( _rhSrcBmp );
	
	// select font
	m_textDC.SelectObject( &m_font );

	// transparent bkg
	m_textDC.SetBkMode( TRANSPARENT );
	
	// draw contour
	m_textDC.SetTextColor( RGB( 0, 0, 0 ) );
	for( int x = -1; x <= 1; x ++ )
		for( int y = -1; y <= 1; y ++ )
			m_textDC.DrawText( _sText, CString( _sText ).GetLength(), CRect( _rctLayout.left + x, _rctLayout.top + y, _rctLayout.right + x, _rctLayout.bottom + y ), _uiFormat | DT_SINGLELINE );

	// draw text
	m_textDC.SetTextColor( RGB( 255, 255, 255 ) );
	m_textDC.DrawText( _sText, CString( _sText ).GetLength(), CRect( _rctLayout ), _uiFormat | DT_SINGLELINE );
}