// A. Sints
#include "stdafx.h"

#include "..\StukaDlg.h"
#include "Script.h"


// script input/group/marker
#include "SInput.h"
#include "SGroup.h"
#include "SMarker.h"


// needed for ffmpeg
#ifndef INT64_C
#define INT64_C(c) (c ## LL)
#define UINT64_C(c) (c ## ULL)
#endif

// FFMPEG capture
#include "FFMPEGCapture.h"

extern "C" {
	#include "libswscale\swscale.h"
}


// dtor
CScript::~CScript() {

	// remove inputs
	CString sID;
	const CSInput * pInput = NULL;
	for( POSITION pos = m_lInput.GetStartPosition(); pos; ) {
		m_lInput.GetNextAssoc( pos, sID, pInput );
		delete pInput;
	}

	// remove groups
	const CSGroup * pGroup = NULL;
	for( POSITION pos = m_lGroup.GetStartPosition(); pos; ) {
		m_lGroup.GetNextAssoc( pos, sID, pGroup );
		delete pGroup;
	}

	// remove markers
	for( POSITION pos = m_lMarker.GetHeadPosition(); pos; )
		delete m_lMarker.GetNext( pos );
}


// parse script
const bool CScript::Parse() {

	// try to read script file
	CFile fScript;
	if( fScript.Open( m_sScriptPath.GetString(), CFile::modeRead ) == FALSE ) {
		CStukaDlg::Output( L"Error (Script): Failed to open script file.\n" );
		return false;
	}

	// read file
	const UINT uiFileLen = UINT( fScript.GetLength() );
	unsigned char * pBuffer = new unsigned char[ uiFileLen ];
	if( fScript.Read( pBuffer, uiFileLen ) != uiFileLen ) {
		fScript.Close();
		CStukaDlg::Output( L"Error (Script): Failed to read script file.\n" );
		return false;
	}

	// put file content in a string
	const CString sScript( pBuffer );
	delete [] pBuffer;

	// close file
	fScript.Close();

	// remove comments
	if( !_RemoveComments( const_cast< CString & >( sScript ) ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to remove comments.\n" );
		return false;
	}

	// parse output
	if( !_ParseOutput( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'output'.\n" );
		return false;
	}
	
	// parse inputs
	if( !_ParseInputs( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'inputs'.\n" );
		return false;
	}

	// parse groups
	if( !_ParseGroups( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'groups'.\n" );
		return false;
	}

	// parse markers
	if( !_ParseMarkers( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'markers'.\n" );
		return false;
	}

	// parse overlay
	if( !_ParseOverlay( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'overlay'.\n" );
		return false;
	}

	// parse range
	if( !_ParseRange( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'range'.\n" );
		return false;
	}

	// parse debug
	if( !_ParseDebug( sScript ) ) {
		CStukaDlg::Output( L"Error (Script): Failed to parse 'debug'.\n" );
		return false;
	}

	// build resize map of markers
	for( POSITION pos = m_lMarker.GetHeadPosition(); pos; ) {
		if( ! const_cast< CSMarker * >( m_lMarker.GetNext( pos ) )->BuildResizeMap() ) {
			CStukaDlg::Output( L"Error (Script): Failed to build resize maps.\n" );
			return false;
		}
	}

	// build speed map of markers
	for( POSITION pos = m_lMarker.GetHeadPosition(); pos; ) {
		if( ! const_cast< CSMarker * >( m_lMarker.GetNext( pos ) )->BuildSpeedMap() ) {
			CStukaDlg::Output( L"Error (Script): Failed to build speed maps.\n" );
			return false;
		}
	}

	// ok
	return true;
}



// validate script
const bool CScript::Validate() {

	// validate inputs
	if( m_lInput.GetCount() == 0 ) {
		CStukaDlg::Output( L"Error (Script): Failed to validate script: 'input' list is empty.\n" );
		return false;
	}

	// validate markers
	if( m_lMarker.GetCount() < 2 ) {
		CStukaDlg::Output( L"Error (Script): Failed to validate script: 'markers' list must have at least two elements.\n" );
		return false;
	}
	if( m_lMarker.GetHead()->GetPosition() != 1 ) {
		CStukaDlg::Output( L"Error (Script): Failed to validate script: first marker must be at measure '1'.\n" );
		return false;
	}

	// validate group elements
	CString sID;
	const CSGroup * pGroup = NULL;
	for( POSITION pos = m_lGroup.GetStartPosition(); pos; ) {
		m_lGroup.GetNextAssoc( pos, sID, pGroup );
		if( !const_cast< CSGroup * >( pGroup )->Validate() ) {
			CStukaDlg::Output( L"Error (Script): Failed to validate script: 'group' validation failed.\n" );
			return false;
		}
		else {

			// validate that groups use known IDs
			CList< CSInput * > lInputs;
			if( !BuildInputList( sID, lInputs ) ) {
				CString sErr;
				sErr.Format( L"Error (Script): Failed to validate script: 'group' ('%s') contains unknown ids.\n", sID.GetString() );
				CStukaDlg::Output( sErr.GetString() );
				return false;
			}
		}
	}

	// validate marker elements
	for( POSITION pos = m_lMarker.GetHeadPosition(); pos; ) {
		const CSMarker * pMarker = m_lMarker.GetNext( pos );
		if( !const_cast< CSMarker * >( pMarker )->Validate() ) {
			CStukaDlg::Output( L"Error (Script): Failed to validate script: 'marker' validation failed.\n" );
			return false;
		}
		else {
			// validate that markers use known IDs
			CList< CString > rRefIDList;
			pMarker->GetRefIDList( rRefIDList );
			for( POSITION pos = rRefIDList.GetHeadPosition(); pos; ) {
				CList< CSInput * > lInputs;
				CString & sID = rRefIDList.GetNext( pos );
				if( !BuildInputList( sID, lInputs ) ) {
					CString sErr;
					sErr.Format( L"Error (Script): Failed to validate script: 'marker' (#%i) contains a reference to unknown id '%s'.\n", pMarker->GetPosition(), sID.GetString() );
					CStukaDlg::Output( sErr.GetString() );
					return false;
				}
			}
		}
	}

	// validate overlay
	if( !m_overlay.Validate() ) {
		CStukaDlg::Output( L"Error (Script): Failed to validate script: 'overlay' validation failed.\n" );
		return false;
	}

	// ok
	return true;
}


// remove all comments
const bool CScript::_RemoveComments( CString & _sScript ) {

	int iStart = 0;
	int iBegin = _sScript.Find( L"//", iStart );
	while( iBegin != -1 ) {
		int iEnd = _sScript.Find( L"\n", iBegin + 1 );
		if( iEnd == -1 ) {
			iEnd = _sScript.GetLength() - 1;
			if( iEnd == -2 ) {
				CStukaDlg::Output( L"Error (Script): Failed to remove comments.\n" );
				return false;
			}
		}
		for( int i = iBegin; i <= iEnd; i++ )
			_sScript.SetAt( i, L' ' );
		iBegin = _sScript.Find( L"//", iStart );
	}

	// ok
	return true;
}


// parse output
const bool CScript::_ParseOutput( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	if( !ExtractCommand( _sScript, L"output", sParams, iStart ) )
		return false;

	int iParamStart = 0;
	if( !ExtractParam( sParams, m_sOutputPath, iParamStart ) )
		return false;
	CString sdwWidth, sdwHeight;
	if( !ExtractParam( sParams, sdwWidth, iParamStart ) )
		return false;
	if( !ExtractParam( sParams, sdwHeight, iParamStart ) )
		return false;
	// optional interpolation type
	CString sInterpolationType( L"fast_bilinear" );
	ExtractParam( sParams, sInterpolationType, iParamStart, true );
	// optional audio path
	ExtractParam( sParams, m_sAudioPath, iParamStart, true );

	// set width/height
	m_dwWidth = _wtoi( sdwWidth.GetString() );
	m_dwHeight = _wtoi( sdwHeight.GetString() );

	// set interpolation type
	CFFMPEGCapture::SetInterpolationType( _GetInterpolationType( sInterpolationType ) );

	// ok
	return true;
}


// get interpolation type from string
int CScript::_GetInterpolationType( const CString & _sInterpolationType ) {

	if( _sInterpolationType == L"fast_bilinear" )	return SWS_FAST_BILINEAR; else
	if( _sInterpolationType == L"bilinear" )		return SWS_BILINEAR; else
	if( _sInterpolationType == L"bicubic" )			return SWS_BICUBIC; else
	if( _sInterpolationType == L"x" )				return SWS_X; else
	if( _sInterpolationType == L"point" )			return SWS_POINT; else
	if( _sInterpolationType == L"area" )			return SWS_AREA; else
	if( _sInterpolationType == L"bicublin" )		return SWS_BICUBLIN; else
	if( _sInterpolationType == L"gauss" )			return SWS_GAUSS; else
	if( _sInterpolationType == L"sinc" )			return SWS_SINC; else
	if( _sInterpolationType == L"lanczos" )			return SWS_LANCZOS; else
	if( _sInterpolationType == L"spline" )			return SWS_SPLINE; else {
		CStukaDlg::Output( L"Warning (Script): Unknown interpolation type '" );
		CStukaDlg::Output( _sInterpolationType );
		CStukaDlg::Output( L"'. defaulting on 'fast_bilinear'\n" );
		return SWS_FAST_BILINEAR;
	}
}


// parse inputs
const bool CScript::_ParseInputs( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	CString sContent;
	while( ExtractBlock( _sScript, L"input", sParams, sContent, iStart ) ) {

		int iParamStart = 0;
		CString sID, sVideoPath;
		CString seFocus( L"center" );
		CString sdwAnchorMs( L"0" ), sdwTLAnchorMs( L"0" );
		if( !ExtractParam( sParams, sID, iParamStart ) )
			return false;
		if( !ExtractParam( sParams, sVideoPath, iParamStart ) )
			return false;
		if( ExtractParam( sParams, seFocus, iParamStart, true ) )
		if( ExtractParam( sParams, sdwAnchorMs, iParamStart, true ) )
			ExtractParam( sParams, sdwTLAnchorMs, iParamStart, true );

		// get focus type
		const CSInput::eFocus eFocus = CSInput::GetFocusType( seFocus );
		if( eFocus == CSInput::IF_NULL )
			return false;

		// create script input
		CSInput * pSInput = new CSInput( sID, sVideoPath, eFocus, _wtoi( sdwAnchorMs.GetString() ), _wtoi( sdwTLAnchorMs.GetString() ) );

		// parse input
		if( !pSInput->Parse( sContent ) ) {
			delete pSInput;
			return false;
		}
		
		// add input to list
		m_lInput.SetAt( sID, pSInput );
	}

	// ok
	return true;
}


// parse groups
const bool CScript::_ParseGroups( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	CString sContent;
	while( ExtractBlock( _sScript, L"group", sParams, sContent, iStart ) ) {

		int iParamStart = 0;
		CString sID;
		if( !ExtractParam( sParams, sID, iParamStart ) )
			return false;

		// create script group
		CSGroup * pSGroup = new CSGroup( sID );

		// parse group
		if( !pSGroup->Parse( sContent ) ) {
			delete pSGroup;
			return false;
		}

		// add group to list
		m_lGroup.SetAt( sID, pSGroup );
	}

	// ok
	return true;
}



// sort by marker position
const bool SortByMarkerPosition( const CSMarker * _pA, const CSMarker * _pB ) {
	return _pA->GetPosition() < _pB->GetPosition();
}



// parse markers
const bool CScript::_ParseMarkers( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	CString sContent;
	while( ExtractBlock( _sScript, L"marker", sParams, sContent, iStart ) ) {

		int iParamStart = 0;
		CString sdwPosition;
		CString seLayout( L"none" );
		if( !ExtractParam( sParams, sdwPosition, iParamStart ) )
			return false;
		ExtractParam( sParams, seLayout, iParamStart, true );

		// get effect type
		const CLayout::eType eLayoutType = CLayout::GetLayoutType( seLayout );
		if( eLayoutType == CLayout::LT_NULL )
			return false;

		// create script marker
		CSMarker * pSMarker = new CSMarker( *this, _wtoi( sdwPosition.GetString() ), eLayoutType );
		
		// parse marker
		if( !pSMarker->Parse( sContent ) ) {
			delete pSMarker;
			return false;
		}

		// add marker to list
		m_lMarker.AddTail( pSMarker );
	}

	// sort the list by position
	m_lMarker.MergeSort( SortByMarkerPosition );

	// set last marker
	const CSMarker * pTailMarker = m_lMarker.GetTail();
	if( pTailMarker != NULL )
		const_cast< CSMarker * >( pTailMarker )->SetTimelineEnd();

	// ok
	return true;
}


// parse overlay
const bool CScript::_ParseOverlay( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	CString sContent;
	while( ExtractBlock( _sScript, L"overlay", sParams, sContent, iStart ) ) {

		int iParamStart = 0;
		CString sInputID, sdwLoopCount;
		if( !ExtractParam( sParams, sInputID, iParamStart ) )
			return false;
		if( !ExtractParam( sParams, sdwLoopCount, iParamStart ) )
			return false;

		// set overlay
		m_overlay.SetOverlay( sInputID, _wtoi( sdwLoopCount.GetString() ) );

		// parse overlay
		if( !m_overlay.Parse( sContent ) )
			return false;
	}

	// ok
	return true;
}


// parse range
const bool CScript::_ParseRange( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	if( ExtractCommand( _sScript, L"range", sParams, iStart ) ) {

		int iParamStart = 0;
		CString sdwRangeBegin, sdwRangeEnd;
		if( !ExtractParam( sParams, sdwRangeBegin, iParamStart ) )
			return false;
		if( !ExtractParam( sParams, sdwRangeEnd, iParamStart ) )
			return false;

		// set range values
		m_dwRangeBegin = _wtoi( sdwRangeBegin.GetString() );
		m_dwRangeEnd = _wtoi( sdwRangeEnd.GetString() );
	}

	// ok
	return true;
}


// parse debug
const bool CScript::_ParseDebug( const CString & _sScript ) {

	int iStart = 0;
	CString sParams;
	if( ExtractCommand( _sScript, L"debug", sParams, iStart ) )
		m_bDebugMode = true;

	// ok
	return true;
}


// build input list from an input or group id
const bool CScript::BuildInputList( const CString & _sRefID, CList< CSInput * > &_rlInputs ) {

	// search in input map
	CMap< CString, LPCTSTR, const CSInput *, const CSInput * >::CPair * pInputPair = m_lInput.PLookup( _sRefID );
	if( pInputPair != NULL ) {
		_rlInputs.AddTail( const_cast< CSInput * >( pInputPair->value ) );
		return true;
	}

	// search in groups
	const CMap< CString, LPCTSTR, const CSGroup *, const CSGroup * >::CPair * pGroupPair = m_lGroup.PLookup( _sRefID );
	if( pGroupPair != NULL ) {
		// recursive call for all group elements
		const CList< CString > & lRefIDList = pGroupPair->value->GetRefIDList();
		for( POSITION pos = lRefIDList.GetHeadPosition(); pos; ) {
			CString sID = lRefIDList.GetNext( pos );
			if( !BuildInputList( sID, _rlInputs ) ) {
				CString sErr;
				sErr.Format( L"Error (Script): Failed to build input list: Unknown ID '%s'.\n", sID.GetString() );
				CStukaDlg::Output( sErr.GetString() );
				return false;
			}
		}
		return true;
	}

	// unknown
	return false;
}